import { by, element, ElementFinder, browser } from "protractor";

export class CalculatorPage {

  getInputLabels() {
    return element.all(by.css("ngparam-input input:not([disabled]) label"));
  }

  getParamInputs() {
    return element.all(by.css("ngparam-input input.form-control"));
  }

  async getParamInputsHavingCalcMode() {
    const ret = [];
    const inputs = this.getParamInputs();
    await inputs.each(async (i) => {
      if (await this.inputHasCalcModeButton(i)) {
        ret.push(i);
      }
    });
    return ret;
  }

  getHeader1() {
    return element(by.css("h1"));
  }

  getSelectById(id: string) {
    return element(by.id(id));
  }

  async getSelectValueText(select: ElementFinder) {
    return await select.element(by.css(".mat-select-value-text > span")).getText();
  }

  async isSelectEmpty(select: ElementFinder) {
    try {
      const text = select.element(by.css(".mat-select-value-text > span"));
      await text.getAttribute("outerHTML"); // await anything trigger the error
      return false;
    } catch (e) { // should be NoSuchElementError
      return true;
    }
  }

  getInputById(id: string) {
    return element(by.css("input#" + id));
  }

  getSaveSessionButton() {
    return element(by.css("dialog-save-session button[type=submit]"));
  }

  getCalculateButton() {
    return element(by.css("button#trigger-calculate"));
  }

  getGeneratePabButton() {
    return element(by.css("button#generate-pab"));
  }

  getCheckedCalcModeButtons() {
    // tslint:disable-next-line:quotemark
    return element.all(by.css('mat-button-toggle.radio_cal[ng-reflect-checked="true"]'));
  }

  getAddStructureButton() {
    return element(by.css("fieldset-container .hyd-window-btns button.add-structure"));
  }

  getAllLinkButtons() {
    return element.all(by.css("mat-button-toggle.radio_link"));
  }

  scrollTo(elt: ElementFinder) {
    browser.controlFlow().execute(function() {
      browser.executeScript("arguments[0].scrollIntoView(true)", elt.getWebElement());
    });
  }

  async inputHasCalcModeButton(input: ElementFinder) {
    // get parent (div.container)
    const container = await this.findParentContainer(input);
    // find radio buttons
    const button: ElementFinder = container.element(by.css("mat-button-toggle.radio_cal > button"));
    return await button.isPresent();
  }

  async hasResults() {
    return (
      await element(by.css("fixedvar-results fixed-results > .fixed-results-container")).isPresent()
      ||
      await element(by.css("fixedvar-results results-graph > graph-results-container")).isPresent()
      ||
      await element(by.css("pab-results pab-results-table")).isPresent()
    );
  }

  async clickSaveCalcButton() {
    return await element(by.css("#save-calc")).click();
  }

  async clickCloneCalcButton() {
    const cloneButton = element(by.css("#clone-calc"));
    return await cloneButton.click();
  }

  async changeSelectValue(elt: ElementFinder, index: number) {
    await elt.click();
    const options = (await elt.getAttribute("aria-owns")).split(" ");
    const optId = options[index];
    const option = element(by.id(optId));
    await option.click();
  }

  // find parent element of elt having class "container"
  async findParentContainer(elt: ElementFinder): Promise<ElementFinder> {
    let i = 8; // garde fous
    while ((await elt.getAttribute("class") !== "container") && (i >= 0)) {
      elt = elt.element(by.xpath(".."));
      i--;
    }
    return elt;
  }

  /**
   * @param elt an <input> element
   * @param mode "fix", "var", "cal" or "link"
   */
  async setParamMode(elt: ElementFinder, mode: string) {
    // get parent (div.container)
    const container = await this.findParentContainer(elt);
    // find radio buttons
    const button = container.element(by.css("mat-button-toggle.radio_" + mode + " > button"));
    await browser.executeScript("window.scrollTo(0, 0);"); // sometimes button slides behind navbar and click() fails
    await button.click();
    // for "var" mode, close the modal
    if (mode === "var") {
      await browser.sleep(500); // wait for the modal to appear
      await element(by.css("dialog-edit-param-values .mat-dialog-actions button")).click();
      await browser.sleep(500); // wait for the navbar to reappear after modal dismissal
    }
  }

  /**
   * @param elt an <input> element
   */
  async getLinkedValueSelect(elt: ElementFinder): Promise<ElementFinder> {
    // get parent (div.container)
    const container = await this.findParentContainer(elt) as ElementFinder;
    // find <select>
    const select = container.element(by.css("param-link mat-select"));
    return select;
  }

  /**
   * Returns an object containing all the calculator's inputs values, indexed
   * by parameter ID
   */
  async storeAllInputValues() {
    const inputs = this.getParamInputs();
    const values = {};
    await inputs.each(async (i) => {
      const inputId = await i.getAttribute("id");
      const inputValue = await i.getAttribute("value");
      values[inputId] = +inputValue; // cast to number to avoid false negative (integers starting with 0)
    });
    return values;
  }

  /**
   * Modifies all the calculator's editable inputs values by adding a random digit other than 0 at the end
   */
  async modifyAllInputValues() {
    const inputs = this.getParamInputs();
    await inputs.each(async (i) => {
      if (await i.isDisplayed()) {
        await i.sendKeys("" + Math.floor(Math.random() * 9) + 1);
      }
    });
  }
}
