import { Component, ViewChild, DoCheck } from "@angular/core";

import { Result, cLog } from "jalhyd";

import { LogComponent } from "../../components/log/log.component";
import { CalculatorResults } from "../../results/calculator-results";
import { NgParameter } from "../../formulaire/ngparam";
import { ApplicationSetupService } from "../../services/app-setup/app-setup.service";
import { PabResultsTableComponent } from "./pab-results-table.component";
import { PabResults } from "../../results/pab-results";
import { PabVariableResultsSelectorComponent } from "./pab-variable-results-selector.component";
import { PlottableData } from "../../results/plottable-data";
import { PlottablePabResults } from "../../results/plottable-pab-results";
import { ResultsGraphComponent } from "../results-graph/results-graph.component";

@Component({
    selector: "pab-results",
    templateUrl: "./pab-results.component.html",
    styleUrls: [
        "../fixedvar-results/fixedvar-results.component.scss"
    ]
})
export class PabResultsComponent implements DoCheck {

    /** résultats non mis en forme */
    private _pabResults: PabResults;

    /** résultats mis en forme pour le graphique */
    private _plottableResults: PlottablePabResults;

    /** true si les résultats doiventt être remis à jour */
    private _doUpdate = false;

    @ViewChild(PabResultsTableComponent)
    private pabResultsTableComponent: PabResultsTableComponent;

    @ViewChild(PabVariableResultsSelectorComponent)
    private pabVariableResultsSelectorComponent: PabVariableResultsSelectorComponent;

    @ViewChild(LogComponent)
    private logComponent: LogComponent;

    @ViewChild(ResultsGraphComponent)
    private resultsGraphComponent: ResultsGraphComponent;

    constructor(
        private appSetupService: ApplicationSetupService,
    ) {
        this._plottableResults = new PlottablePabResults();
    }

    public set results(rs: CalculatorResults[]) {
        this._pabResults = undefined;
        if (rs.length > 0 && rs[0] instanceof PabResults) {
            this._pabResults = rs[0] as PabResults;
        }
        this.updateView();
    }

    /**
     * update results table and chart when the variable index changed (event sent by
     * PabVariableResultsSelectorComponent); variable index is already set in
     * pabResults at this time
     */
    public variableIndexChanged() {
        console.log("oulala l'index variableu qui change !!");
        this.updateView();
    }

    public updateView() {
        this.logComponent.log = undefined;
        if (this.pabResultsTableComponent) {
            this.pabResultsTableComponent.results = undefined;
        }
        if (this.pabVariableResultsSelectorComponent) {
            this.pabVariableResultsSelectorComponent.results = undefined;
        }
        if (this.resultsGraphComponent) {
            this.resultsGraphComponent.results = undefined;
        }
        // set _doUpdate flag so that results are rebuilt on the next Angular display cycle
        this._doUpdate = false;
        if (this._pabResults !== undefined) {
            this._doUpdate = this._doUpdate || this._pabResults.hasResults || this._pabResults.hasLog;
        }
    }

    public ngDoCheck() {
        if (this._doUpdate) {
            this._doUpdate = !this.updateResults();
        }
    }

    private mergeLog(result: Result, log: cLog) {
        if (result && result.hasLog) {
            if (result.hasGlobalLog) {
                log.addLog(result.globalLog);
            } else {
                log.addLog(result.log);
            }
        }
    }

    private get mergedGlobalLogs(): cLog {
        const res = new cLog();
        if (this._pabResults) {
            this.mergeLog(this._pabResults.result, res);
            // log des cloisons
            for (const c of this._pabResults.cloisonsResults) {
                this.mergeLog(c, res);
            }
            // log de la cloison aval
            this.mergeLog(this._pabResults.cloisonAvalResults, res);
        }
        return res;
    }

    /**
     * met à jour l'affichage des résultats
     * @returns true si les résultats ont pu être mis à jour
     */
    private updateResults() {
        let pabUpdated: boolean;
        let graphUpdated: boolean;
        if (this.hasResults) {
            pabUpdated = this.pabResultsTableComponent !== undefined;
            if (pabUpdated) {
                this.pabResultsTableComponent.results = this._pabResults;
                this.pabVariableResultsSelectorComponent.results = this._pabResults;
            }
            graphUpdated = this.resultsGraphComponent !== undefined;
            if (graphUpdated) {
                this.resultsGraphComponent.results = this.plottableResults;
                this.resultsGraphComponent.updateView();
            }
        } else {
            pabUpdated = true;
            graphUpdated = true;
        }

        const logUpdated = this.logComponent !== undefined;
        if (logUpdated) {
            this.logComponent.log = this.mergedGlobalLogs;
        }

        return pabUpdated && logUpdated && graphUpdated;
    }

    public get pabResults() {
        return this._pabResults;
    }

    public formattedLabel(p: NgParameter): string {
        return CalculatorResults.paramLabel(p, false);
    }

    public formattedValue(p: NgParameter): string {
        const nDigits = this.appSetupService.displayDigits;
        return p.getValue().toFixed(nDigits);
    }

    public get hasResults(): boolean {
        return this._pabResults && this._pabResults.hasResults;
    }

    /** builds a set of PlottableData from PabResults, to feed the graph */
    protected get plottableResults(): PlottableData {
        this._plottableResults.setPabResults(this.pabResults);
        return this._plottableResults;
    }
}
