import { PlottableData } from "./plottable-data";
import { PabResults } from "./pab-results";
import { GraphType } from "./graph-type";
import { ServiceFactory } from "../services/service-factory";

export class PlottablePabResults implements PlottableData {

    public graphType: GraphType = GraphType.Scatter;
    public chartX: string;
    public chartY: string;

    protected pabResults: PabResults;

    public constructor(pabResults?: PabResults) {
        if (pabResults) {
            this.setPabResults(pabResults);
        }
        // axes par défaut
        this.chartX = this.chartX || "CLOISON";
        this.chartY = this.chartY || "Z";
    }

    /** reaffect pabResults, for ex. when objet was contructed with empty pabResults */
    public setPabResults(pabResults: PabResults) {
        this.pabResults = pabResults;
    }

    /**
     * Returns the label to display, for an element of getAvailableChartAxis()
     * @param symbol parameter / result symbol (ex: "Q")
     */
    public getChartAxisLabel(symbol: string): string {
        if (symbol === "x") { // specific case for wall abscissa
            return ServiceFactory.instance.i18nService.localizeText("INFO_LIB_ABSCISSE_CLOISON");
        } else {
            return this.pabResults.headers[this.pabResults.columns.indexOf(symbol)];
        }
    }

    public expandLabelFromSymbol(symbol: string): string {
        return symbol;
    }

    /**
     * Returns a list of plottable parameters / result elements, that can be defined
     * as X or Y chart axis
     */
    public getAvailableChartAxis(): string[] {
        // add wall abscissa on the fly
        return [ "x" ].concat(this.pabResults.columns);
    }

    // just to implement interface
    public getVariatingParametersSymbols(): string[] {
        return [];
    }

    /**
     * Returns the series of values for the required symbol
     * @param symbol parameter / result symbol (ex: "Q")
     */
    public getValuesSeries(symbol: string): any[] {
        const data: string[] = [];
        const pr = this.pabResults;
        const nDigits = ServiceFactory.instance.applicationSetupService.displayDigits;
        const l = this.pabResults.cloisonsResults.length;
        // when a parameter is variating, index of the variating parameter
        // values to build the data from
        const vi = this.pabResults.variableIndex;

        if (this.hasError()) {
            return [];
        }

        switch (symbol) {
            case "CLOISON":
                data.push("");
                for (let i = 0; i <= l; i++) { // <= for one extra step (downwall)
                    data.push("" + (i + 1));
                }
                break;

            case "DH":
            case "ZRAM":
            case "Q":
                data.push("");
                for (let i = 0; i < l; i++) {
                    const er = pr.cloisonsResults[i].resultElements[vi].getExtraResult(symbol);
                    data.push((er !== undefined) ? er.toFixed(nDigits) : "");
                }
                const zrAval = pr.cloisonAvalResults.resultElements[vi].getExtraResult(symbol);
                data.push((zrAval !== undefined) ? zrAval.toFixed(nDigits) : "");
                break;

            case "Z":
                for (let i = 0; i < l; i++) {
                    data.push(pr.cloisonsResults[i].resultElements[vi].vCalc.toFixed(nDigits));
                }
                data.push(pr.cloisonAvalResults.resultElements[vi].vCalc.toFixed(nDigits));
                data.push(pr.Z2.toFixed(nDigits));
                break;

            case "PV":
            case "YMOY":
            case "ZRMB":
            case "QA":
                data.push("");
                for (let i = 0; i < l; i++) {
                    const er = pr.cloisonsResults[i].resultElements[vi].getExtraResult(symbol);
                    data.push((er !== undefined) ? er.toFixed(nDigits) : "");
                }
                data.push("");
                break;

            case "x": // wall abscissa
                data.push("");
                for (let i = 0; i < l; i++) {
                    const er = pr.cloisonsResults[i].resultElements[vi].getExtraResult(symbol);
                    data.push((er !== undefined) ? er.toFixed(nDigits) : "");
                }
                const erXdw = pr.cloisonAvalResults.resultElements[vi].getExtraResult(symbol);
                data.push((erXdw !== undefined) ? erXdw.toFixed(nDigits) : "");
                break;
        }

        return data;
    }

    protected hasError(): boolean {
        return this.pabResults.hasError();
    }
}
